/* ***** BEGIN LICENSE BLOCK *****
 * Licensed under Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * Full Terms at http://mozile.mozdev.org/license.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Playsophy code.
 *
 * The Initial Developer of the Original Code is Playsophy
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *		James A. Overton <james@overton.ca>
 *
 * ***** END LICENSE BLOCK ***** */

/**
@fileoverview

These functions handle the behaviour of the Source Edit box.
*/

/**
Constructs a MozileSource object.

@constructor
*/
function MozileSource()
{
	// XUL element shortcuts
	this.mozileSourceSplitter = null;  // Splits the window.
	this.mozileSourceBox = null;       // Main source view container.
	this.mozileSourceTextbox = null;   // Source view textbox.
	this.mozileSourceSync = null;      // "synchronise" checkbox.
	this.mozileSourceNamespace = null; // "namespace" checkbox.
	this.mozileSourceError = null;     // source error button.

	this.mode = 2;                     // source editing mode: 1=page, 2=block
	this.parseError = null;            // XML document containing the last parser error message
	this.lastEA = null;
	
	this.target = null;
	this.selection = null;
	
	// Constants
	this.PAGE = 1;
	this.BLOCK = 2;
}

/**
If synchronised, updates the browser view from the source (markup) view.
*/
MozileSource.prototype.sourceInput = function()
{
	if( this.isSourceBox() && this.isSync() )
	{
		this.updateBrowserView();
	}
};

/**
If synchronised, updates the source (markup) view from the browser view.
This should be called after a formatting command that changes the browser view.
*/
MozileSource.prototype.browserInput = function(target,selection)
{
	if( this.isSourceBox() && this.isSync() )
	{
		this.updateSourceView(target,selection);
	}
};


/**
Toggles between block and page source modes.

@param {String} mode "page" or "block".
*/
MozileSource.prototype.switchMode = function( mode, target, selection )
{
	if ( mode == "page" ) {
		this.mode = this.PAGE;
	} else {
		this.mode = this.BLOCK;
	}
	this.updateSourceView(target,selection);
};

/**
Pops up a window with the details of the last XML parser error.
*/
MozileSource.prototype.showParseError = function()
{
	var p = this.parseError.getElementsByTagName('parsererror')[0].firstChild.nodeValue;
	var s = this.parseError.getElementsByTagName('sourcetext')[0].firstChild.nodeValue;
	p = strToHtml(p);
	s = strToHtml(s);

	var doc = 
'<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">\n'+
'<html xmlns="http://www.w3.org/1999/xhtml">\n'+
'<head><title>Mozile Error</title></head>\n'+
'<body style="margin: 0px; padding: 0px 20px 2px 4px; color: #000; background-color: #fff;">\n'+
'<h4 style="margin-top: 1px; white-space: pre;">'+p+'</h4>\n'+
'<pre style="margin-bottom: 1px; color: #339;">'+s+'</pre>\n'+
'</body>\n'+
'</html>\n';
	
	var win = window.open("","mozileErrorWindow", "chrome,resizable=yes,scrollbars=yes,minimizable=yes");
	win.document.open();
	win.document.write(doc);
	win.document.close();

	function strToHtml(s)
	{
		var h = s.replace(/&/g,'&amp;');
		h = h.replace(/</g,'&lt;');
		h = h.replace(/>/g,'&gt;');
		return h;
	}
};

/**
Tests if the source box is present and visible.

@return isSourceBox true if present and visible, otherwise false.
@type Boolean
*/
MozileSource.prototype.isSourceBox = function()
{
	if(!this.mozileSourceBox){
		this.mozileSourceBox = document.getElementById("mozileSourceBox");
		if (this.mozileSourceBox) {
			this.mozileSourceTextbox = document.getElementById("mozileSourceTextbox");
			this.mozileSourceNamespace = document.getElementById("mozileSourceNamespace");
			this.mozileSourceSync = document.getElementById("mozileSourceSync");
			this.mozileSourceError = document.getElementById('mozileSourceError');
			this.mozileSourceSplitter = document.getElementById("mozileSourceSplitter");
		}
	}	
	if( (this.mozileSourceBox) && (this.mozileSourceBox.style.visibility != "collapse") ) {
		return true;
	} else {
		return false;	
	}
};

MozileSource.prototype.isNamespace = function()
{
	if ( (this.mozileSourceNamespace) && (this.mozileSourceNamespace.getAttribute("checked") == "true") ) {
		return true;
	}
	return false;
};

MozileSource.prototype.isSync = function()
{
	if ( (this.mozileSourceSync) && (this.mozileSourceSync.getAttribute("checked") == "true") ) {
		return true;
	}
	return false;
};

MozileSource.prototype.showErrorButton = function()
{
	if ( this.mozileSourceError ) {
		this.mozileSourceError.setAttribute("collapsed","false");
	}
};

MozileSource.prototype.hideErrorButton = function()
{
	if ( this.mozileSourceError ) {
		this.mozileSourceError.setAttribute("collapsed","true");
	}
};

/**
Updates the source (xhtml/xml markup) view.
*/
MozileSource.prototype.updateSourceView = function(target,selection)
{
	this.target=target;
	this.selection=selection;

	if (!this.isSourceBox()) {return;}
	// If the source box is not visible no need to do anything.

	var f = ["mozileSourceEdit.js","updateSourceView"];
	mozile.bug.write(f,3,"Source view is being updated...");

	try {
		if(this.target) {
			var editableArea="";
			try {
				if(this.mode==this.BLOCK) { //Edit current block
					var cssr = this.selection.getEditableRange();
					if (cssr) {
						editableArea = cssr.top;
					} else {
						editableArea = this.lastEA;
					}
				}
				else { //Edit whole page
					editableArea = this.target.document.documentElement;
				}
				this.lastEA=editableArea;
				mozile.bug.write(f,2,"Editable Area: "+editableArea);
			}
			catch(e)
			//Catch errors if mozileSelection or eDOM are no longer valid.
			{
				mozile.bug.write(f,0,"No Editable Area: "+e);
				editableArea = this.lastEA;
//				this.lastEA = null;
			}

			if (!editableArea) { // this can happen while eDOM loads...
				mozile.bug.write(f,0,"No Editable Area");
				//clear target/selection etc?
				return;
			}

			// Decide whether to use HTML or XML mode
			var xml;
			if(this.target.document.contentType=="text/html")
			{
				xml = mozileXhtmlCompatible(mozileHTML2xhtml(mozileSerializeCurrent(editableArea)));
			}
			else
			{
				if( this.isNamespace() ) {
					xml = mozileSerializeCurrentNS(editableArea);
				}
				else {
					xml = mozileSerializeCurrent(editableArea);
				}
			}
			// XML Serializer only converts <>&.
			// Must manually convert non-break space otherwise textbox converts them to plain spaces!
//			xml = mozileConvertToEntities(xml, 7);
			xml = xml.replace(/\xA0/g,'&nbsp;');
			this.mozileSourceTextbox.value = xml;
		}
		else {
			this.mozileSourceTextbox.value =  mozile.strings.getString("se.no.selection");
			this.target=null;
			this.selection=null;
			this.lastEA = null;
			mozile.bug.write(f,2,"No editable selection.");
		}
	}
	catch(e)
	{
		mozile.bug.write(f,0,"Error updating source view: "+e);
		//Don't overwrite the source box.
	}
};

/**
Updates the browser view.
To do: add in save to Undo stack.

@return result true if browser view has changed (or Exception and may have changed); false if parsing error; null if nothing done or Exception.
@type Boolean 
*/
MozileSource.prototype.updateBrowserView = function()
{
	var f = ["mozileSourceEdit.js","updateBrowserView"];
	mozile.bug.write(f,3,"Browser view is being updated...");
	
	var result;

	if (!this.isSourceBox()) {return null;}

	var editableArea="";
	try
	{
		if (this.target) {
			if(this.mode==this.PAGE) { // edit current block.
				editableArea = this.target.document.documentElement;
			}
			else { // edit the whole page
				var cssr = this.selection.getEditableRange();
				if (cssr) {
					editableArea = cssr.top;
				} else {
					editableArea = this.lastEA;
				}
			}
			this.lastEA = editableArea;
			mozile.bug.write(f,3,"Editable Area: "+editableArea);
		}
		else {
			// Is this right?
			editableArea=this.lastEA=null;
		}
	}
	catch(e)
	{
		mozile.bug.write(f,0,"No Editable Area: "+e);
		editableArea=this.lastEA=null;
	}

	if (editableArea)
	{
		result = this.setInnerXML(editableArea, this.mozileSourceTextbox.value);
		if(result===false) {
			this.showErrorButton();
		}
		else { //Ok or Exception...
			this.hideErrorButton();
			result=true;
		}
	}
	else
	{
		mozile.bug.write(f,2,"No editable area to update.");
		result=null;
	}
	return result;
};

/**
Converts <code>str</code> into a DOM fragment and replaces the contents of <code>node</code>
with the DOM fragment.

<p>The XML/XHTML parser is used for all documents.

<p>The xhtml namespace attribute is added to the root element of documents of MIME type text/html
if they don't already have it. [The alternative is to use innerHTML, but then you have to treat the
head and body sections separately.] 

<p>Global variable <code>mozileSourceError</code> contains the parsing error info if there is an error.

@param {Element} node
@param {String} str
@return true if done ok, false if error when parsing "str", null if there is an exception.
@type Boolean. 
*/
MozileSource.prototype.setInnerXML = function(node, str)
{
	var f = ["mozileSourceEdit.js","setInnerXML"];
	mozile.bug.write(f,3,"Setting inner XML...");

	try
	{
		var namespace, xmlNamespace;
		var contentType = node.ownerDocument.contentType;

		switch ( contentType )
		{
			case "text/html" :
				contentType = "text/xml";
				namespace = "http://www.w3.org/1999/xhtml";
				// If text/html make sure the xhtml namespace is set for the root element.
				if ( node.ownerDocument.documentElement.getAttribute( "xmlns" ) != namespace ) {
					node.ownerDocument.documentElement.setAttribute( "xmlns", namespace );
				}
				break;
			/*
			DOM parser can use application/xhtml+xml MIME type according to the docs...
			...BUT xhtml entity parsing is NOT supported. Can only use text/xml.
			*/
			default :
				contentType = "text/xml";
				namespace = node.namespaceURI;
				break;
		}	

		if (namespace) {
			xmlNamespace = ' xmlns="'+namespace+'"';
		} else {
			xmlNamespace = "";
		}
		var wrapperNodeName = node.localName;
		// Must convert &nbsp; to unicode before passing to the XML parser.
		var xmlStr = str.replace(/&nbsp;/g,'\xA0');
		xmlStr = "<"+wrapperNodeName+xmlNamespace+">"+xmlStr+"<\/"+wrapperNodeName+">";
		
		var p	=	new DOMParser();
		var docFrag = p.parseFromString(xmlStr,contentType);
		
		if(docFrag.documentElement.nodeName == "parsererror")
		{
			mozile.bug.write(f,2,"Parser Error!");
			this.parseError = docFrag;
			return false;
		}
		else
		{
			while (node.hasChildNodes()) {
				node.removeChild(node.lastChild);
			}
			var nodeX = docFrag.getElementsByTagName(wrapperNodeName)[0];
			while (nodeX.hasChildNodes()) {
				node.appendChild(nodeX.firstChild);
			}
			return true;
		}
	}
	catch(e)
	{
		mozile.bug.write(f,0,"Exception: "+e);
		return null;
	}
};

/**
Toggles the visibility of the source edit box.
*/
MozileSource.prototype.toggleSourceView = function(target,selection)
{
	var f = ["mozileSourceEdit.js","showSourceView"];
	mozile.bug.write(f,3,"Toggling source view...");

	if( !this.mozileSourceBox ) {
		// If source box not present, need to load it.
		mozile.bug.write(f,3,"Creating source edit tools");

		// Get the container
		var mozileSourceVbox = document.getElementById("appcontent");
		
		var xmlDoc = this.loadSourceBox();
		if (xmlDoc) {
			var sourceSplitter = xmlDoc.getElementById("mozileSourceSplitter").cloneNode(true);
			var sourceBox = xmlDoc.getElementById("mozileSourceBox").cloneNode(true);
			
			// append the XUL nodes
			mozileSourceVbox.appendChild(sourceSplitter);
			mozileSourceVbox.appendChild(sourceBox);

			this.isSourceBox(); //Set shortcuts.
			this.mozileSourceTextbox.value =  mozile.strings.getString("se.no.selection");
			this.showSourceView();
			return;
		}
	}

	if( this.mozileSourceBox ) {
		// If I use the "collapse" XUL attribute then the textbox won't trigger oninput events, so instead we use the "visibility" CSS property
		if(this.mozileSourceBox.style.visibility == "collapse") {
			this.showSourceView();
			this.browserInput(target,selection); //If in sync, update source view.
		}
		else {
			this.hideSourceView();
		}
	}
	
};

/**
Hides the source editing features if they are loaded.
*/
MozileSource.prototype.hideSourceView = function()
{
	if( this.mozileSourceBox )
	{
		this.mozileSourceSplitter.style.visibility = "collapse";
		this.mozileSourceBox.style.visibility = "collapse";
	}
};

/**
Shows the source editing features if they are loaded.
*/
MozileSource.prototype.showSourceView = function()
{
	if( this.mozileSourceBox )
	{
		this.mozileSourceSplitter.style.visibility = "visible";
		this.mozileSourceBox.style.visibility = "visible";
	}
};

/**
Gets the sourcebox.xul file as an XUL document.
*/
MozileSource.prototype.loadSourceBox = function()
{
	var f = ["mozileSourceEdit.js","loadSourceBox"];
	mozile.bug.write(f,3,"Load Source Box starting");
	
	var sourceBoxFile = "chrome://mozile/content/XUL/sourcebox.xul";
	var	loaded, xmlDoc;
	
	try {
 		xmlDoc = document.implementation.createDocument("", "", null);
		xmlDoc.async = false;
		loaded = xmlDoc.load(sourceBoxFile);
		if( !loaded ) {
			mozile.bug.write(f,1,"Source Box file load has failed");
			return false;
		}
	}catch(e){
		mozile.bug.write(f,1,"Exception: "+e);
		return false;
	}

	mozile.bug.write(f,3,"Source Box file loaded");
	return xmlDoc;
};
