/* ***** BEGIN LICENSE BLOCK *****
 * Licensed under Version: MPL 1.1/GPL 2.0/LGPL 2.1
 * Full Terms at http://mozile.mozdev.org/license.html
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Playsophy code.
 *
 * The Initial Developer of the Original Code is Playsophy
 * Portions created by the Initial Developer are Copyright (C) 2002-2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 * Max d'Ayala, June 2005 - new module created.
 *
 * ***** END LICENSE BLOCK ***** */

/**
Constructor for special chars object.

@constructor
*/
function MozileSpecial( entry, specials ) {
	this.entry = entry || "any";
	this.selectedIndex = 0;
	this.specials = [];
	this.add = mozileSpecialAdd;
	this.clear = mozileSpecialClear;
}

/**
Clear special char blocks.

@member MozileSpecial
*/
function mozileSpecialClear() {
	this.specials = [];
}

/**
Add special char blocks.

@param {Array} blocks array of blocks to add.
@member MozileSpecial
*/
function mozileSpecialAdd( blocks ) {
	this.specials = this.specials.concat( blocks );
}


/**
Constructor for special chars block.

@param {string} name identifier
@param {integer} columns number of columns in the character grid.
@param {object} specialblocks array of character/range configuration data.
@param {integer} selectedIndex initial block to select/display.
@constructor
*/
function MozileSpecialBlock( name, columns, specialblocks, selectedIndex ) {
	this.name = name;
	this.columns = columns || 10 ;
	this.specialblocks = new Array();
	if (specialblocks)
		this.specialblocks = this.specialblocks.concat(specialblocks);
	this.selectedIndex = selectedIndex || 0;
}

/* Initialisation on startup. */
/* Global variables */

const MOZILE_SPECIAL_LATIN1 = [{range:{start:160, end:255, wrap:25}}];
const MOZILE_SPECIAL_XHTML_SYMBOLS = [
	{characters:[169,174,"S",36,162,163,8364,402,165,164]},  //Currency
	{characters:[176,8216,8217,8220,8221,171,187,8230,161,191,166,167,182]}, //Punctuation, general
	{characters:[188,189,190,185,178,179,"S",960]}, //Numbers
	{characters:[215,247,183,8730,8733,8734,177,8773,8776,8800,8804,8805,8756,8747]}, //Basic Maths
	{characters:[8592,8593,8594,8595,8596,8656,8657,8658,8659,8660]}, //Arrows
];
const MOZILE_SPECIAL_XHTML_ACCENTS = [
	{characters:
		[192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,
		 216,338,352,217,218,219,220,221,222,376,223]},
	{characters:
		[224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,240,241,242,243,244,245,246,
		 248,339,353,249,250,251,252,253,254,255]}
];
const MOZILE_SPECIAL_XHTML_GREEK = [
	{characters:[913,914,915,916,917,918,919,920,921,922,923,924,925,926,927,928,929,931,"S",932,933,934,935,936,937]},
	{characters:[945,946,947,948,949,950,951,952,953,954,955,956,957,958,959,960,961,963,962,964,965,966,967,968,969]},
	{characters:[977,978,982]}
];

/* Global functions */

/**
Initialise the Special Characters Object to the default values.
*/
function mozileSpecialInit( specialObj ) {
	var blocks = [
		new MozileSpecialBlock( "XHTML Symbols", 25, MOZILE_SPECIAL_XHTML_SYMBOLS ),
		new MozileSpecialBlock( "XHTML Accents", 25, MOZILE_SPECIAL_XHTML_ACCENTS ),
		new MozileSpecialBlock( "XHTML Greek", 25, MOZILE_SPECIAL_XHTML_GREEK ),
		new MozileSpecialBlock( "Latin1", 25, MOZILE_SPECIAL_LATIN1 )
	];
	specialObj.clear();
	specialObj.add( blocks );
}


/**
Open Special Chars window and set up unload event handler.
*/
function mozileSpecialOpen( target ) {
	var win = window.openDialog("chrome://mozile/content/XUL/special.xul",
		"special", "chrome,modal,left=0,top=0,resizable=no", target.document.mozileSpecial);
}

/**
Parse the Special Character data in the MES document.

@param target window to apply special char data to.
@param MES xml configuration file.

@return true - special chars found and processed ok, false - no special chars to process, null - error.
*/
function mozileParseSpecialChars(target, MES) {
	var f = ["mozileSpecialChars.js","mozileParseSpecialChars"];
	mozile.bug.write(f,3,"Looking for Special characters in MES.");
	var specialchars;
	
	try {
		// Initialise defaults.
		target.document.mozileSpecial = new MozileSpecial();
		mozileSpecialInit(target.document.mozileSpecial);
		
		specialchars = MES.documentElement.getElementsByTagName("specialchars");
		if (specialchars.length > 0) {
			if ( mozileParseSpecials(specialchars[0], target.document.mozileSpecial) ) {
				mozile.bug.write(f,2,"MES Special characters parsed.");
				return true;
			} else {
				mozile.bug.write(f,1,"Parsing of special chars in MES failed.");
				return null;		
			}
		} else {
			mozile.bug.write(f,3,"No special character definitions in MES.");
			return false;
		}
	} catch (e) {
		mozile.bug.write(f,1,"Exception: "+e);
		return null;
	}
}

/* Local functions */

/**
Parse special characters element.

@param {Node} specialchars element node.
@param {object} specialObj to store parsed data into.
@return true - parse completed ok; null - error.
*/
function mozileParseSpecials(specialchars, specialObj) {
	var f = ["mozileSpecialChars.js","mozileParseSpecials"];
	var i,j, block, blocks, entry, insert, specialentry, specialblock;

	try {
		// optional, defaults to insert="add"
		if ( specialchars.hasAttribute("insert") ) {
			insert = specialchars.getAttribute("insert");
			insert = insert.toLowerCase();
			switch ( insert ) {
				case "add" :
				case "replace" :
					break;
				default :
					insert="add"
			}
		} else {
			insert="add";
		}

		// <specialentry> tag is optional. Defaults to entry="any"
		specialentry =  specialchars.getElementsByTagName("specialentry");
		if ( specialentry.length > 0 && specialentry[0].hasAttribute("entry") ) {
			entry = specialentry[0].getAttribute("entry");
			entry = entry.toLowerCase();
			switch ( entry ) {
				case "any" :
				case "none" :
					break;
				default :
					entry="any";
			}
		} else {  // default: allow entry of any value.
			entry = "any";
		}

		// Get the data from all the <specialblock>'s. (Optional)
		blocks = new Array();
		specialblock = specialchars.getElementsByTagName("specialblock");
		if ( (j=specialblock.length) > 0 ) {
			if ( j > 30 ) j=30;
			for( i=0; i<j; i++ ) {
				block = mozileParseSpecialblock( specialblock[i], i );
				if(  block === null )
					mozile.bug.write(f,1,"specialblock["+i+"] parse error.");
				else
					blocks.push( block );
			}
		}
		
		// Save the special data.
		specialObj.entry = entry;
		if ( insert == "replace" )
			specialObj.clear();
		specialObj.add( blocks );
		return true;

	} catch (e) {
		mozile.bug.write(f,1,"Error: "+e);
		return null;
	}
}

// Return: blockObj
//
function mozileParseSpecialblock(specialblock, n) {
	var f = ["mozileSpecialChars.js","mozileParseSpecialblock"];
	var name, columns, colsAttr, blockData, block, chars, range, blockObj;

	try {
		if ( specialblock.hasAttribute("name") ) {
			name = specialblock.getAttribute("name");
			name = name.substr(0,30);
		} else {
			mozile.bug.write(f,2,"Missing name attribute" );
			name = "Special "+n;
		}
			
		if ( specialblock.hasAttribute("columns") ) {
			colsAttr = specialblock.getAttribute("columns");
			columns = parseInt( colsAttr );
			if (isNaN( columns ) || (columns == 0) ) {
				mozile.bug.write(f,1,"Invalid value: columns=\""+colsAttr+'"' );
				columns = 10;
			}
			if ( columns > 50 )
				columns = 50;
		} else {
			columns = 10;
		}
		
		blockData = new Array();
		block = specialblock.childNodes;
		for (var i=0; i<block.length; i++) {
			if (block[i].nodeType == Node.ELEMENT_NODE) {
				switch ( block[i].nodeName ) {
					case "characters" :
						chars = mozileParseCharacters(block[i]);
						if ( chars !== null )
							blockData.push( {"characters":chars} );
						break;
					case "range" :
						range = mozileParseRange(block[i], columns);
						if ( range !== null )
							blockData.push( {"range":range} );
						break;
					//ignore anything else.
				}
			}
		}
		if ( blockData.length == 0 )
			throw "Missing block data.";
		blockObj = new MozileSpecialBlock( name, columns, blockData );
		return blockObj;
				
	} catch (e) {
		mozile.bug.write(f,1,"Error: "+e);
		return null;
	}
}

//	Parse a character element.
// 	Get the data from an element such as:
//	<characters>160,161,...,255</characters>
//	Whitespace is stripped from the data. 
//	The comma separated list is split and returned as an array of values.
//
function mozileParseCharacters(characters) {
	var f = ["mozileSpecialChars.js","mozileParseCharacters"];
	var charData, re;
	
	try {
		if ( !characters.hasChildNodes() || (characters.firstChild.nodeType != Node.TEXT_NODE) )
			throw "Missing <characters> data";
		charData = characters.firstChild.nodeValue;
		charData = charData.substr(0,1000);
		re = /\s/g;
		charData = charData.replace( re, "" );	//remove whitespace.
		charData = charData.split(",");	//split comma separated list into an array.
		return charData;
	} catch (e) {
		mozile.bug.write(f,3,"Error: "+e);
		return null;
	}
}

function mozileParseRange( range, columns ) {
	var f = ["mozileSpecialChars.js","mozileParseRange"];
	var rangeObj, start, end, wrap;
	
	try {
		if ( range.hasAttribute("start") ) {
			start = range.getAttribute("start");
			start = parseInt( start );
			if (isNaN( start ))
				throw "\start\" is not a number";
		} else {
			throw "Required \"start\" attribute is missing";
		}

		if ( range.hasAttribute("end") ) {
			end = range.getAttribute("end");
			end = parseInt( end );
			if (isNaN( end ))
				throw "\end\" is not a number";
		} else {
			throw "Required \"end\" attribute is missing";
		}

		// "wrap" attribute is optional.
		if ( range.hasAttribute("wrap") ) {
			wrap = range.getAttribute("wrap");
			wrap = parseInt( wrap );
			if (isNaN( wrap )) {
				mozile.bug.write(f,1,"Error: wrap attribute has invalid value.");
				wrap = columns;
			}
			if ( wrap > columns )
				wrap = columns;
		} else {
			wrap = columns;
		}

		rangeObj = new Object();
		rangeObj.start=start;
		rangeObj.end=end;
		rangeObj.wrap=wrap;
		return rangeObj;

	} catch (e) {
		mozile.bug.write(f,1,"Error: "+e);
		return null;
	}
}